# Duo, A crazy eights game.
# Copyright (C) 2003-2008 The Duo developers.
#
# This file is part of Duo.
#
# Duo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""The protocol used between Duo players.

The Duo protocol is the way for a player to communicate with other
players (users or AIs) and the referee. The protocol defines messages
for the every game event. Some of these messages can only be emitted
by the referee, some only by the players. Similarly, some messages
must only be addressed to the referee, and some to the players. At
last, some messages must be sent to every player, some others only to
one player.

Note that the protocol does not cover the media used by communication. This
is the task of the transport module.

Hereafter, C will stand for client, R for referee and A for all
players.

Authentication
--------------
There is no authentication at the Duo protocol level. Authentication
must be achieved by the underlying trannsport layer. It can ensure for
example that two players always have a different name, or that there
is no mismatch on connections.

Everytime a player joins a game, the game referee broadcasts an
Authentication message to all the player of the game. This messages
provides the updated list of the game players.

  R->A: AUTH <player_name_list>

This way, all players are informed of the new comer. This one also gets the name
of all the previous players connected and the one of the Referee.

Giving cards
------------
The game starts when a player send the Start message with the correct
game key. The referee then deals the cards to the players.

  P->R: START <key>
  R->C: HAND <card_list>

Every player can take the time to read them before signaling he is ready.
This message is broadcasted because players can be interested to be
informed in real time of players status.

  C->A: RDY

When all the players are ready, the referee returns the first card.
If the first card is a wild card, it chooses the color.

  R->A: HEAP <card, color, player>

Playing a card
--------------

  A player can make a move with:
  C->R: PLAY <card, color>

The referee checks the move to refuse or accept it:

  R->A: ERROR <card, player>
  R->A: TURN <player>

or

  R->C: HAND <card_list>
  R->A: HEAP <card, color, player>
  R->A: TURN <player>

Drawing a card
--------------

  C->R: DRAW
  R->C: HAND <card_list>
  R->A: DEAL <player, nb_cards>

Passing
-------

  C->R: PASS
  R->A: TURN <player>

"""

import logging

from duo import transport


class Message(object):
    """The Duo protocol piece of information.

    A Duo message is implicitely defined by its type, and the
    information it carries is completed by its payload. The payload
    vary with the kind of information carried by the message. If more
    than one object is attached to the message, the payload is a
    tuple.
    """
    def __init__(self, payload=None):
        self._payload = payload

    def __str__(self):
        return self.__class__.__name__ + ': ' + str(self._payload)

    def __repr__(self):
        return self.__class__.__name__ + ': ' + str(self._payload)

    def _get_payload(self):
        """Retrieve the content of this message."""
        return self._payload

    payload = property(_get_payload, None, None, 'The Message payload')


# From referee
class AuthMessage(Message):
    """Message broadcasted by the referee each time a player joins the game.

    This message payload is the new players list.
    """
    def __init__(self, players_list):
        Message.__init__(self, players_list)


class DealMessage(Message):
    """Message broadcasted by the referee when a player receives new cards.

    This message is sent to warn all the players. Its payload is a
    tuple with the number of cards drawn and the recipient name. The
    player who drew the cards receives the Hand message which provides
    his full hand.
    """
    def __init__(self, nb_cards, player):
        Message.__init__(self, (nb_cards, player))


class HandMessage(Message):
    """Message sent by the referee to a player who drew one or more cards.

    This message payload is the new full hand of the player.
    """
    def __init__(self, hand):
        Message.__init__(self, hand)


class HeapMessage(Message):
    """Message broadcasted by the referee each time the heap changes.

    The message payload is a tuple with the new heap, the demanded
    color, and the player who played ths card.
    """
    def __init__(self, card, color, player):
        Message.__init__(self, (card, color, player))


class TurnMessage(Message):
    """Message broadcasted by the referee each time a new player should play.

    The message payload is the name of the player who should now play.
    """
    def __init__(self, player):
        Message.__init__(self, player)


class ReverseMessage(Message):
    """Message broadcasted by the referee when a Reverse card is played.

    There is no payload attached to this message.
    """
    def __init__(self):
        Message.__init__(self)


class WinnerMessage(Message):
    """Message broadcasted by the referee when the game came to an end.

    The message paylod is the name of the winner.
    """
    def __init__(self, player):
        Message.__init__(self, player)


# From players
class StartMessage(Message):
    """Message emitted by a player to start the game.

    The player must provide a key, which will be checked by the
    referee. The game will only start if the provided key matches the
    key used to create the referee.
    """
    def __init__(self, key):
        Message.__init__(self, key)


class ReadyMessage(Message):
    """Message emitted by players when they are ready to start.

    Once a player receives his cards, he should have the time to
    inspect his hand, sort the cards, or whatever his UI allows him to
    do. This message should be emitted when the player feels ready. It
    must be automatically sent by an AI when it receives its cards.

    The referee will wait all the players are ready prior to draw the
    first card and put it on the heap.

    The is no payload attached to this message.
    """
    def __init__(self):
        Message.__init__(self)


class PlayMessage(Message):
    """Message emitted by a player to play a card.

    The message payload is the card played, and the color asked.
    """
    def __init__(self, card, color):
        Message.__init__(self, (card, color))


class DrawMessage(Message):
    """Message emitted by a player to draw.

    There is no payload attached to this message. The number of cards
    to be drawn depends on the context and is decided by the referee.
    """
    def __init__(self):
        Message.__init__(self)


class PassMessage(Message):
    """Message emitted by a player to pass.

    There is no payload attached to this message.
    """
    def __init__(self):
        Message.__init__(self)


# Any way
class ErrorMessage(Message):
    """Message emitted by an entity on error.

    The message payload is the exception raised, or the error
    string. Currently, this message is only used by the referee.
    """
    def __init__(self, exception):
        Message.__init__(self, exception)


class MessageBundle(list):
    """A group of message."""
    pass


class PlayerAlreadyAuthenticatedError(Exception):
    """Error raised on multiple player authentication."""
    def __init__(self):
        Exception.__init__(self)


class ProtocolAdapter(object):
    """The interface to the transport layer dedicated to the Duo protocol.

    The ProtocolAdapter provides basic registering to the transport
    layer, and Duo messages mangling and demangling.
    """
    def __init__(self, name):
        self._name = name
        self._hooks = {}
        self._endpoint = None

    def connect(self, transport_layer, transport_options):
        """Connect to the transport layer."""
        transport_options['callbacks'] = {
            transport.TransportJoinPacket: self.on_endpoint_join,
            transport.TransportEscapePacket: self.on_endpoint_escape,
            transport.ProtocolPacket: self.on_message }

        self._endpoint = transport.create_endpoint(
            transport_layer, transport_options, self.name)

    def is_connected(self):
        """Return true is connected, falsee otherwise."""
        return self._endpoint != None

    def register(self, msg_class, callback, argument=None):
        """Register a new callback."""
        msg_type = msg_class.__name__
        logging.getLogger('duo.protocol').debug(
            '%s registered %s with %s' % (self.name, type, callback.__name__))
        self._hooks[msg_type] = (callback, argument)

    def on_message(self, message, emitter):
        """Callback invoked by the transport layer on new message."""
        if isinstance(message, MessageBundle):
            for msg in message:
                self._notify(msg, emitter)
        else:
            self._notify(message, emitter)

    def send_to(self, message, recipient):
        """Send a message to one recipient."""
        logging.getLogger('duo.protocol').debug(
            '%s sends %s to %s' % (self.name, message, recipient))
        self._endpoint.send_to(message, recipient)

    def broadcast(self, message):
        """Send a message to all the recipients in the same context."""
        logging.getLogger('duo.protocol').debug(
            '%s broadcasts: %s' % (self.name, message))
        self._endpoint.broadcast(message)

    def on_endpoint_join(self, message, emitter):
        """Callback invoked by the transport layer on new connection."""
        pass

    def on_endpoint_escape(self, message, emitter):
        """Callback invoked by the transport layer on disconnection."""
        pass

    def _get_name(self):
        """Retrieve the name of this adapter."""
        return self._name

    name = property(_get_name, None, None, "The Protocol Adapter name")

    def _notify(self, message, emitter):
        """Dispatch the reveiced message to the registered callbacks."""
        msg_type = message.__class__.__name__
        try:
            callback, arg = self._hooks[msg_type]
            logging.getLogger('duo.protocol').debug(
                'notifying %s to %s with %s'
                % (msg_type, self.name, callback.__name__))
            callback.__call__(message.payload, emitter, arg)
        except KeyError, ex:
            logging.getLogger('duo.protocol').debug(
                'ProtocolAdapter._notify no callback for %s in %s' %
                (message.__class__.__name__, ex))
        except Exception, ex:
            logging.getLogger('duo.protocol').error(
                'ProtocolAdapter._notify %s: %s' %
                (message.__class__.__name__, ex.__class__.__name__))


class Player(ProtocolAdapter):
    """The player class from a user point of view.

    Any user interface needs to work with at least one Player instance;
    the one representing the user for the game. There should also be one
    instance for each one of the AI players running on the machine.

    The Player interface provides the minimal but strictly necessary
    methods to manage a player representation.
    """
    def __init__(self, name, cb_argument=None, ai=None):
        ProtocolAdapter.__init__(self, name)

        self._hand = []
        self._referee = None
        self._ai = ai
        if self._ai:
            self._ai.initialize(self)

        # Connect essential callbacks for a player.
        self.register(AuthMessage, self.on_auth_cb, cb_argument)
        self.register(HandMessage, self.on_hand_cb, cb_argument)
        self.register(TurnMessage, self.on_turn_cb, cb_argument)
        self.register(HeapMessage, self.on_heap_cb, cb_argument)
        self.register(DealMessage, self.on_deal_cb, cb_argument)
        self.register(ErrorMessage, self.on_error_cb, cb_argument)
        self.register(WinnerMessage, self.on_winner_cb, cb_argument)

    def is_robot(self):
        """Return True if this player is an automaton."""
        return self._ai != None

    def send_to_referee(self, message):
        """Send the given message to the referee only."""
        self.send_to(message, self._referee)

    def get_games(self):
        """Retrieve the available games in the neighbourhood."""
        return self._endpoint.get_contexts()

    def join_game(self, name):
        """Join the game whose name is given."""
        self._endpoint.join_context(name)

    def leave_game(self):
        """Leave the current game."""
        self._endpoint.escape()
        if self._ai:
            self._ai.stop_game()

    def _get_hand(self):
        """Retrieve the current cards of the player."""
        return self._hand

    def _set_hand(self, cards):
        """Update the cards of the player."""
        self._hand = cards

    hand = property(_get_hand, _set_hand, None, 'The player hand')

    def start_game(self, key=None):
        """Ask the referee to deal the cards."""
        if self._ai:
            self._ai.start_game()
        else:
            self.send_to_referee(StartMessage(key))

    def ready(self):
        """Get ready to start playing."""
        self.broadcast(ReadyMessage())

    def play(self, card, color):
        """Play a card."""
        self.send_to_referee(PlayMessage(card, color))

    def draw(self):
        """Draw a given number of cards."""
        self.send_to_referee(DrawMessage())

    def pass_turn(self):
        """Pass the turn."""
        self.send_to_referee(PassMessage())

    def on_auth_cb(self, payload, emitter, args):
        """Hook called on authentication."""
        self._referee = emitter
        if self._ai:
            self._ai.on_auth(payload, emitter, args)

    def on_hand_cb(self, payload, emitter, args):
        """Hook called on new hand."""
        self._hand = payload
        if self._ai:
            self._ai.on_hand(payload, emitter, args)

    def on_turn_cb(self, payload, emitter, args):
        """Hook called on new turn."""
        if self._ai:
            self._ai.on_turn(payload, emitter, args)

    def on_heap_cb(self, payload, emitter, args):
        """Hook called on new heap."""
        if self._ai:
            self._ai.on_heap(payload, emitter, args)

    def on_deal_cb(self, payload, emitter, args):
        """Hook called when a new deal is done, whoever the player."""
        if self._ai:
            self._ai.on_deal(payload, emitter, args)

    def on_error_cb(self, payload, emitter, args):
        """Hook called on error."""
        if self._ai:
            self._ai.on_error(payload, emitter, args)

    def on_winner_cb(self, payload, emitter, args):
        """Hook called on game end."""
        if self._ai:
            self._ai.on_winner(payload, emitter, args)
