# Duo, A crazy eights game.
# Copyright (C) 2003-2008 The Duo developers.
#
# This file is part of Duo.
#
# Duo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Duo user interfaces.

"""

import os
import sys
import time

from duo import tools


# Boot options.
OPT_GAME_PLAYER_NAME = 'name'
OPT_GAME_TRANSPORT = 'transport'
OPT_GAME_NB_PLAYERS = 'players'
OPT_GAME_NB_CARDS = 'cards'
OPT_GAME_START = 'start'

def sanitize_options(raw_options):
    options = raw_options
    if options is None:
        options = {}
    if not options.has_key(OPT_GAME_PLAYER_NAME):
        options[OPT_GAME_PLAYER_NAME] = 'Player'
    if not options.has_key(OPT_GAME_TRANSPORT):
        options[OPT_GAME_TRANSPORT] = 'local'
    if not options.has_key(OPT_GAME_NB_PLAYERS):
        options[OPT_GAME_NB_PLAYERS] = 4
    if not options.has_key(OPT_GAME_NB_CARDS):
        options[OPT_GAME_NB_CARDS] = 7
    if not options.has_key(OPT_GAME_START):
        options[OPT_GAME_START] = False
    return options


def get_uis():
    """Get a list of all the available UIs."""
    path = os.path.dirname(sys.modules['duo.ui'].__file__)
    modules = filter(
        lambda x: x != '__init__.py' and not x.endswith('.pyc'),
        os.listdir(path))
    return [os.path.splitext(file_[0]) for file_ in modules]


def create_ui(ui, options):
    """Instantiate the UI with the given name."""
    module = tools.import_recursively('duo.ui.' + ui)
    return module.UserInterface(options)

#
# Here are reusable service objects.
#
class Game(object):
    """TODO: model as in MVC.
    """
    def __init__(self, name):
        self._name = name
        self._key = time.time()
        self._referee = None

        # Keeps the *ordered* list of players.
        self._player_list = []

        # Player dictionary.
        # key: player name,
        # value: tuple (player, score)
        self._players = {}

    def _get_name(self):
        """Return the name of this game."""
        return self._name

    name = property(_get_name, None, None, 'The game name')

    def _get_key(self):
        """Return the key of this game.

        A game can only be started using its key.
        """
        return self._key

    key = property(_get_key, None, None, 'The game key')

    def _get_referee(self):
        """Get the referee associated to this game."""
        return self._referee

    def _set_referee(self, referee):
        """Set the referee associated to this game."""
        self._referee = referee

    referee = property(_get_referee, _set_referee, None, 'The game referee')

    def add_player(self, player):
        """Add a new player to the game."""
        self._player_list.append(player)
        self._players[player.name] = (player, 0)

    def get_players(self):
        """Retrieve the list of the players in this game."""
        return self._player_list

    def get_player(self, name):
        """Retrieve the player with the given name."""
        try:
            return self._players[name][0]
        except KeyError:
            return None

    def get_score(self, name):
        """Retrieve the score for the given player name."""
        try:
            return self._players[name][1]
        except KeyError:
            return -1

    def add_to_score(self, name, score):
        """Update the score of a player."""
        try:
            tuple_ = self._players[name]
            self._players[name] = (tuple_[0], tuple_[1] + score)
        except KeyError:
            pass

    def end(self):
        """Conclude the game.

        All players and the referee are disconnected.
        """
        for player in self._player_list:
            # Leave the game, then leave again to disconnect.
            player.leave_game()
            player.leave_game()
        self._referee.leave()
