# Duo, A crazy eights game.
# Copyright (C) 2003-2008 The Duo developers.
#
# This file is part of Duo.
#
# Duo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


"""
Commands:

Settings:

get transports : get available transport layers
get robots : get available AI
get rulesets : get the defined rulesets.

set transport <layer_name>
set player <player_name>
set bot <bot_name>
set ruleset -- some day :)

Game:

start
ready
draw
pass
xx: a card to play

bye
"""

import curses
import curses.ascii

import logging
import os
import socket
import time

from duo import protocol
from duo import transport
from duo import game
from duo import referee
from duo import ai


class PlayerUI(protocol.Player):
    def __init__(self, name, transport_layer, transport_options, ui):
        protocol.Player.__init__(self, name, transport_layer, transport_options)
        self._ui = ui
        self._distant_players = []

    def set_ui(self, ui):
        self._ui = ui

    def get_distant_players(self):
        return self._distant_players

    def on_auth_cb(self, message, emitter, args):
        protocol.Player.on_auth_cb(self, message, emitter, args)
        self._distant_players = []
        for player in message.get_payload():
            if player != self.get_name() :
                self._distant_players.append(player)

    def on_hand_cb(self, message, emitter, args):
        protocol.Player.on_hand_cb(self, message, emitter, args)
        if self._ui:
            self._ui.display_hand()

    def on_heap_cb(self, message, emitter, args):
        """TODO"""
        protocol.Player.on_heap_cb(self, message, emitter, args)
        if self._ui:
            card, color, player = message.get_payload()
            self._ui.display_heap(card, color, player)

    def on_error_cb(self, message, emitter, args):
        protocol.Player.on_error_cb(self, message, emitter, args)
        if self._ui:
            self._ui.display_error(str(message.get_payload()))

    def on_turn_cb(self, message, emitter, args):
        protocol.Player.on_turn_cb(self, message, emitter, args)
        if self._ui:
            self._ui.display_turn(message.get_payload())

    def on_pass_cb(self, message, emitter, args):
        protocol.Player.on_pass_cb(self, message, emitter, args)
        #print '## ' + aMessage.getPayload() + ' passes'

    def on_reverse_cb(self, message, emitter, args):
        protocol.Player.on_reverse_cb(self, message, emitter, args)
        if self._ui:
            self._ui.display_reverse_event()


class GameWindow(object):
    """The actual game view.

    The window is divided into three parts. First, the heap, then the
    players hands, and at the bottom of the screen, the status line
    which provides information on the game.
    """

    INTERVAL = 2
    HEAP_LINE = 2
    PLAYER_LINE = 4
    OFFSET = 2

    def __init__(self, interface, screen):
        self._ui = interface
        self._screen = screen
        self._cursor = 0
        self._message_queue = []

        self._player = interface.get_player()
        self._player.set_ui(self)
        self._players = [self._player.name]
        self._players.extend(self._player.get_distant_players())

        self._draw()

        logging.getLogger('duo.ui.dreadful').debug('START')
        self._player.start_game(self._ui.get_game_key())

    def run(self):
        stop_flag = False
        while not stop_flag:
            c = self._wnd.getch()
            if c in [ord('r'), ord('R')]:
                self._player.ready()

            elif c in [ord(' '), curses.KEY_ENTER, 10]:
                card = self._player.get_hand()[self._cursor]
                if card.color == 'N':
                    # Select color
                    pass
                self._player.play(card, card.color)

            elif c in [ord('p'), ord('P'), curses.KEY_LEFT]:
                self._draw_cursor(self._cursor, delete=True)
                self._cursor = self._cursor - 1
                self._constrain_cursor()
                self._draw_cursor(self._cursor)

            elif c in [ord('n'), ord('N'), curses.KEY_RIGHT]:
                self._draw_cursor(self._cursor, delete=True)
                self._cursor = self._cursor + 1
                self._constrain_cursor()
                self._draw_cursor(self._cursor)

            elif c in [ord('d'), ord('D')]:
                self._player.draw()

            elif c in [ord('q'), ord('Q')]:
                stop_flag = True

            self._wnd.refresh()

    def display_hand(self):
        self._draw_cards(0, self._player.get_hand())
        self._constrain_cursor()
        self._draw_cursor(self._cursor)
        self._wnd.refresh()

    def display_heap(self, card, color, player):
        self._draw_heap(card, color)
        self._draw_message('%s played %s(%s)' % (player, card, color))
        self._wnd.refresh()

    def display_turn(self, player):
        if player != self._player.get_name():
            self._draw_message('It\'s %s turn' % player)
        else:
            self._draw_message('It\'s your turn!')
        self._wnd.refresh()

    def display_error(self, error):
        self._draw_message('ERROR: %s' % error)
        self._wnd.refresh()

    def _constrain_cursor(self):
        if self._cursor < 0:
            self._cursor = 0;
        elif self._cursor >= len(self._player.get_hand()):
            self._cursor = len(self._player.get_hand()) - 1

    def _draw_cursor(self, pos, delete=False):
        self._wnd.move(GameWindow.PLAYER_LINE,
                       GameWindow.OFFSET + len(self._player.name) + 3 * pos - 1)
        if delete:
            self._wnd.addch(' ')
        else:
            self._wnd.addch('<')
        self._wnd.move(GameWindow.PLAYER_LINE,
                       GameWindow.OFFSET + len(self._player.name) + 3 * pos + 2)
        if delete:
            self._wnd.addch(' ')
        else:
            self._wnd.addch('>')

    def _draw_heap(self, card, color):
        self._wnd.move(GameWindow.HEAP_LINE, GameWindow.OFFSET)
        self._wnd.addstr('On the heap: ')
        self._wnd.addch(ord(card.type))
        self._wnd.addch(ord(card.color))
        self._wnd.addstr(' (%s)' % color)

    def _draw_cards(self, player_idx, hand):
        line = GameWindow.PLAYER_LINE + player_idx * GameWindow.INTERVAL

        # First reset the line.
        self._wnd.move(line, GameWindow.OFFSET)
        y, x = self._wnd.getmaxyx()
        self._wnd.addstr(' ' * (x - 2 * GameWindow.OFFSET))

        # Then redraw it.
        self._wnd.move(line, GameWindow.OFFSET)
        self._wnd.addstr(self._players[player_idx])
        for idx in range(len(hand)):
            self._wnd.move(line,
                           GameWindow.OFFSET + len(self._player.name) + 3 * idx)
            self._wnd.addch(ord(hand[idx].type))
            self._wnd.addch(ord(hand[idx].color))

    def _draw_message(self, message):
        """Update the bottom message box.

        The messages scroll on three lines.
        """
        self._message_queue.append(message)
        if len(self._message_queue) > 3:
            self._message_queue.pop(0)

        y, x = self._wnd.getmaxyx()
        for i in range(len(self._message_queue)):
            self._wnd.move(y - GameWindow.OFFSET - i, GameWindow.OFFSET)
            self._wnd.addstr(self._message_queue[-i - 1] + ' '
                             * (x - len(self._messageQueue[-i - 1]) - 2 * GameWindow.OFFSET))

    def _draw(self):
        (height, width) = self._screen.getmaxyx()
        curses.curs_set(0)
        self._wnd = self._screen.subwin(height, width, 0, 0)
        self._wnd.box()

        self._draw_heap(game.Card('_', '_'), '_')

        for i in range(len(self._players)):
            self._drawCards(i, [game.Card('_', '_')] * 7)
        self._wnd.refresh()


class BoardWindow(object):
    STATUS_OK = 0
    STATUS_ERROR = 1
    STATUS_GAME_START = 10
    STATUS_QUIT = 20

    UNARY_COMMANDS = ['help', 'start', 'connect', 'quit']
    BINARY_COMMANDS = ['get', 'set']

    def __init__(self, interface, screen):
        self._ui = interface

        (height, width) = screen.getmaxyx()
        self._wnd = screen.subwin(height, width, 0, 0)
        self._wnd.refresh()

    def run(self):
        status = BoardWindow.STATUS_OK
        line = ''
        self.print_string('Duo dreadful interface. (Fear!)')
        while not status in [BoardWindow.STATUS_GAME_START, BoardWindow.STATUS_QUIT] :
            c = self._wnd.getch()
            if c in [curses.KEY_BACKSPACE, 127]:
                if len(line):
                    line = line[:-1]
                    self.delete_character()

            elif c in [curses.KEY_ENTER, 10]:
                tokens = line.lower().split()
                if tokens[0] in BoardWindow.BINARY_COMMANDS and len(tokens) >= 2:
                    # Try to call the matching callback.
                    cb = 'on' + tokens[0].capitalize() + tokens[1].capitalize()
                    if cb in UserInterface.__dict__:
                        logging.getLogger('duo.ui.dreadful').debug(
                            'invoking binary command %s' % cb)
                        (status, message) = UserInterface.__dict__[cb](self._ui, tokens[2:])

                elif tokens[0] in BoardWindow.UNARY_COMMANDS:
                    # Try to call the matching callback.
                    cb = 'on' + tokens[0].capitalize()
                    if cb in UserInterface.__dict__:
                        logging.getLogger('duo.ui.dreadful').debug(
                            'invoking unary command %s' % cb)
                        (status, message) = UserInterface.__dict__[cb](self._ui, tokens[1:])

                else:
                    (status, message) = (BoardWindow.STATUS_ERROR, 'Unknown command')

                line = ''
                self.print_string(message)

            elif curses.ascii.isalnum(c) or curses.ascii.isspace(c):
                line += chr(c)
                self.print_character(c)

        return status

    def print_character(self, char):
        self._wnd.echochar(char)

    def delete_character(self):
        (y, x) = self._wnd.getyx()
        self._wnd.move(y, x - 1)
        self._wnd.addch(' ')
        self._wnd.move(y, x - 1)
        self._wnd.refresh()

    def print_string(self, blob):
        (y, x) = self._wnd.getyx()
        self._wnd.move(y + 1, 0)
        if blob is None:
            self._wnd.addstr('')
        elif (type(blob) is list):
            for item in blob:
                self._wnd.addstr(item + ' ')
        else:
                self._wnd.addstr(blob)
        self._wnd.move(y + 2, 0)
        self._wnd.addstr('> ')
        self._wnd.cursyncup()
        self._wnd.refresh()


# Callbacks of console return a tuple : (status, message)
# where message will be displayed and status is defined in
# BoardWindow
class UserInterface(object):
    SCREEN_BOARD = 'board'
    SCREEN_HELP  = 'help'
    SCREEN_GAME  = 'game'

    HAND_WINDOW_HEIGHT = 4

    def __init__(self):
        self._available_layers = []
        self._available_games= []
        self._current_layer = ''

        self._game = None
        self._game_key = time.time()

        self._local_player_name = ' '.join(
            [str(os.getpid()), 'on', socket.gethostname()])
        self._local_player = None
        self._distant_players = None
        self._local_bots = []

        self._init_screen()
        logging.getLogger('duo.ui.dreadful').info('dreadful initialized')

    def get_player(self):
        return self._local_player

    def get_game_key(self):
        return self._game_key

    def launch(self):
        """TODO"""
        stop_flag = False
        try:
            while not stop_flag:
                wnd = self._draw_screen(UserInterface.SCREEN_BOARD)
                status = wnd.run()
                if status == BoardWindow.STATUS_GAME_START:
                    wnd = self._draw_screen(UserInterface.SCREEN_GAME)
                    wnd.run()
                else:
                    stop_flag = True
        finally:
            self._finalize_screen()

    def on_set_transport(self, tokens):
        """TODO"""
        self._fetch_transport_layers()
        status = BoardWindow.STATUS_OK
        message = ''

        # TODO: concatenate tokens.
        layer = tokens[0]
        result = (BoardWindow.STATUS_OK, '')
        if layer in self._available_layers:
            if layer != self._current_layer:
                self._current_layer = layer
                # TODO: Use OptionSets
                transport_options = { 'context': 'root' }
                self._local_player = PlayerUI(self._local_player_name, self._current_layer, transport_options, self)
                self._bots = []
                message = 'Layer %s chosen.' % self._current_layer
            else:
                message = 'Layer unchanged.'
        else:
            (status, message) = (BoardWindow.STATUS_ERROR, '%s is not a valid layer.' % layer)
        return (status, message)

    def on_get_transports(self, tokens):
        """TODO"""
        self._fetch_transport_layers()
        return (BoardWindow.STATUS_OK, self._available_layers)

    def on_set_game(self, tokens):
        """TODO"""
        status = BoardWindow.STATUS_OK
        message = 'No connection available.'
        try:
            if self._local_player:
                self._game = self._local_player.get_name() + '_party'
                # TODO: Use OptionSets
                transport_options = { 'context': self._game }
                referee_options = {
                    'game': self._game,
                    'cards': 7,
                    'ruleset': 'DEFAULT_RULESET',
                    'master_key': self._game_key
                }

                self._referee = referee.Referee(
                    self._current_layer, transport_options, referee_options)
                message = 'created %s' % self._game
                self._local_player.join_game(self._game)

        except transport.BadContext:
            status = BoardWindow.STATUS_ERROR
            message = 'Not in root context.'
        return (status, message)

    def on_get_games(self, tokens):
        """TODO"""
        status = BoardWindow.STATUS_ERROR
        message = 'No connection available.'
        try:
            self._fetch_games()
            if self._local_player:
                status = BoardWindow.STATUS_OK
                message = self._local_player.get_games()
        except transport.BadContext:
            message = 'Not in root context.'
        return (status, message)

    def on_set_player(self, tokens):
        """TODO"""
        self._local_player_name = ' '.join(tokens)
        return (BoardWindow.STATUS_OK, 'Set player name to %s' % self._local_player_name)

    def on_set_bot(self, tokens):
        """TODO"""
        status = BoardWindow.STATUS_ERROR
        message = 'No connection available.'
        try:
            if self._game:
                # TODO: Use OptionSets
                transport_options = { 'context': self._game }
                bot = ai.load(aTokens[0], self._current_layer, transport_options)
                self._local_bots.append(bot)
                status = BoardWindow.STATUS_OK
                message = ('Bot ' + bot.get_name() + ' added')
        except Exception, e:
            message = '!!' + str(e)
        return (status, message)

    def on_start(self, tokens):
        return (BoardWindow.STATUS_GAME_START, '')

    def _fetch_transport_layers(self):
        """TODO"""
        if len(self._available_layers) == 0:
            logging.getLogger('duo.ui.dreadful').debug(
                'fetching available transport layers')
            self._available_layers = transport.get_layers()

    def _fetch_games(self):
        """TODO"""
        logging.getLogger('duo.ui.dreadful').debug(
            'fetching available games')
        if self._local_player:
            self._available_games = self._local_player.get_games()

    def _init_screen(self):
        self._screen = curses.initscr()
        try:
            curses.noecho()
            curses.cbreak()
            self._screen.keypad(1)
        except Exception:
            self._finalize_screen()
            raise

    def _draw_screen(self, screen):
        try:
            self._screen.erase()
            if screen == UserInterface.SCREEN_BOARD:
                return BoardWindow(self, self._screen)

            elif screen == UserInterface.SCREEN_HELP:
                (heigth, width) = self._screen.getmaxyx()
                return self._screen.subwin(10, width, (heigth - 10) / 2, 0)

            elif screen == UserInterface.SCREEN_GAME:
                return GameWindow(self, self._screen)

        except Exception:
            self._finalize_screen()
            raise



    def _finalize_screen(self):
        curses.nocbreak()
        self._screen.keypad(0)
        curses.echo()
        curses.endwin()
